#pragma once

#include <MultiThreading/Thread.h>
#include <vector>
#include <queue>
namespace MultiThreading {
    class ThreadPool;
}

class MultiThreading::ThreadPool {
private:
    std::vector<MultiThreading::Thread*> threads;
    std::queue<std::shared_ptr<TaskBase>> queue;
    std::mutex queue_mutex;
private:
    /// @returns nullptr if the queue is empty.
    std::shared_ptr<TaskBase> Dequeue();
    void Runner(const std::shared_ptr<TaskBase>& task);
public:
    void Enqueue(const std::shared_ptr<MultiThreading::TaskBase>& task);
    void Enqueue(const std::function<void()>& task);
public:
    [[nodiscard]] unsigned int ThreadCount() const { return threads.size(); }
    [[nodiscard]] unsigned int QueueSize();
public:
    ThreadPool();
    explicit ThreadPool(unsigned int thread_count);
    ~ThreadPool();
};

/// Canonical C++ ThreadPool object
/// Ref. https://stackoverflow.com/questions/15752659/thread-pooling-in-c11
class ThreadPool {
public:
    void Start()
    {
        const uint32_t num_threads = std::thread::hardware_concurrency(); // Max # of threads the system supports.
        for (uint32_t ii = 0; ii < num_threads; ++ii) {
            threads.emplace_back(&ThreadPool::ThreadLoop, this);
        }
    }
    void QueueJob(const std::function<void()>& job)
    {
        {
            std::unique_lock<std::mutex> lock(queue_mutex);
            jobs.push(job);
        }
        mutex_condition.notify_one();
    }
    void Stop() {
        {
            std::unique_lock<std::mutex> lock(queue_mutex);
            should_terminate = true;
        }
        mutex_condition.notify_all();
        for (std::thread& active_thread : threads) {
            active_thread.join();
        }
        threads.clear();
    }
    bool busy() {
        bool bussy;
        {
            std::unique_lock<std::mutex> lock(queue_mutex);
            bussy = !jobs.empty();
        }
        return bussy;
    }
private:
    void ThreadLoop()
    {
        while (true)
        {
            std::function<void()> job;
            {
                std::unique_lock<std::mutex> lock(queue_mutex);
                mutex_condition.wait(lock, [this] {
                   return !jobs.empty() || should_terminate;
                });

                if (should_terminate) { return; }
                job = jobs.front();
                jobs.pop();
            }
            job();
        }
    }

    bool should_terminate = false;
    std::mutex queue_mutex;
    std::condition_variable mutex_condition;
    std::vector<std::thread> threads;
    std::queue<std::function<void()>> jobs;
};